import re
from os.path import isfile, join
from registry import REGISTRY_CODE


def registry_lines(pkgs_dir, info):
    pat = re.compile(r'([\w.]+)\s+([\w.]+-\d+)$')
    for rs in ['%(name)s %(version)s-%(build)d' % info] + info['packages']:
        m = pat.match(rs)
        if not m:
            print "Warning: not a full requirement:", rs
            continue
        pkg = '%s-%s' % (m.group(1).lower(), m.group(2))
        reg_path = join(pkgs_dir, pkg, 'EGG-INFO', 'registry.txt')
        if not isfile(reg_path):
            #print "Warning: no registry file:", reg_path
            continue
        for line in open(reg_path):
            line = line.strip()
            if line.startswith('#'):
                yield line
            else:
                k, v = line.split(None, 1)
                assert v.startswith('../'), v
                yield '%s  ../../%s/%s' % (k, pkg, v[3:])


def create_entry_script(path, entry):
    """
    create entry point Python script at 'path', which reads app_registry.txt
    """
    assert entry.count(':') == 1
    module, func = entry.strip().split(':')
    fo = open(path, 'w')
    fo.write(REGISTRY_CODE)
    fo.write("""
if __name__ == '__main__':
    from os.path import isfile
    path = join(dirname(__file__), 'app_registry.txt')
    if isfile(path):
        update_registry([path])
    else:
        print "Warning: no registry file:", path
    from %(module)s import %(func)s
    sys.exit(%(func)s())
""" % locals())
    fo.close()


def create_entry(egg, info):
    if egg.hook:
        with open(join(egg.meta_dir, 'app_registry.txt'), 'w') as fo:
            for line in registry_lines(egg.pkgs_dir, info):
                fo.write('%s\n' % line)

    app_entry = info.get('app_entry')
    if app_entry:
        create_entry_script(join(egg.meta_dir, 'app_entry.py'), app_entry)



import os
import re
import json
import string
from os.path import abspath, getsize, getmtime, isdir, isfile, join

from utils import comparable_version, info_file
from egg_meta import is_valid_eggname, split_eggname
try:
    import zdiff
except ImportError:
    zdiff = None


fn_pat = re.compile(r'([\w.]+)-([\w.]+)-(\d+)--([\w.]+)-(\d+)\.zdiff$')
def split(fn):
    m = fn_pat.match(fn)
    if m is None:
        raise Exception("Can not split: %r" % fn)
    return m.expand(r'\1-\2-\3.egg'), m.expand(r'\1-\4-\5.egg')


def update_patches(eggs_dir, patches_dir, verbose=False):

    def calculate_all_patches():
        egg_names = [fn for fn in os.listdir(eggs_dir)
                     if is_valid_eggname(fn)]
        names = set(split_eggname(egg_name)[0]
                    for egg_name in egg_names)
        for name in sorted(names, key=string.lower):
            versions = []
            for egg_name in egg_names:
                n, v, b = split_eggname(egg_name)
                if n != name:
                    continue
                versions.append((v, b))
            versions.sort(key=(lambda vb: (comparable_version(vb[0]), vb[1])))
            versions = ['%s-%d' % vb for vb in versions]
            lv = len(versions)
            #print name, lv, versions
            for i in xrange(0, lv):
                for j in xrange(i + 1, lv):
                    yield '%s-%s--%s.zdiff' % (name, versions[i], versions[j])

    def up_to_date(patch_fn):
        patch_path = join(patches_dir, patch_fn)
        if not isfile(patch_path):
            return False
        info = zdiff.info(patch_path)
        for t in 'dst', 'src':
            if getmtime(join(eggs_dir, info[t])) != info[t + '_mtime']:
                return False
        return True

    def create(patch_fn):
        src_fn, dst_fn = split(patch_fn)
        src_path = join(eggs_dir, src_fn)
        dst_path = join(eggs_dir, dst_fn)
        assert isfile(src_path) and isfile(dst_path)
        if verbose:
            print 'creating', patch_fn
        patch_path = join(patches_dir, patch_fn)
        zdiff.diff(src_path, dst_path, patch_path + '.part')
        os.rename(patch_path + '.part', patch_path)

    all_patches = set()
    for patch_fn in calculate_all_patches():
        all_patches.add(patch_fn)
        if not up_to_date(patch_fn):
            create(patch_fn)

    # remove old patches
    for patch_fn in os.listdir(patches_dir):
        if patch_fn.endswith('.zdiff') and patch_fn not in all_patches:
            os.unlink(join(patches_dir, patch_fn))


def update_index(eggs_dir, patches_dir, force=False):
    index_path = join(patches_dir, 'index.json')
    if force or not isfile(index_path):
        index = {}
    else:
        index = json.load(open(index_path))

    new_index = {}
    for patch_fn in os.listdir(patches_dir):
        if not fn_pat.match(patch_fn):
            continue
        src_fn, dst_fn = split(patch_fn)
        dst_path = join(eggs_dir, dst_fn)
        dst_size = getsize(dst_path)
        if dst_size < 131072:
            continue
        patch_path = join(patches_dir, patch_fn)
        if dst_size < getsize(patch_path) * 2:
            continue
        info = index.get(patch_fn)
        if info and getmtime(patch_path) == info['mtime']:
            new_index[patch_fn] = info
            continue
        info = info_file(patch_path)
        info.update(zdiff.info(patch_path))
        info['name'] = patch_fn.split('-')[0].lower()
        new_index[patch_fn] = info

    with open(index_path, 'w') as f:
        json.dump(new_index, f, indent=2, sort_keys=True)


def update(eggs_dir, force=False, verbose=False):
    if zdiff is None:
        print "Warning: could not import bsdiff4, cannot create patches"
        return

    patches_dir = join(eggs_dir, 'patches')
    if not isdir(patches_dir):
        os.mkdir(patches_dir)

    if force:
        index_files = ['index.json']
        for fn in os.listdir(patches_dir):
            if fn.endswith('.zdiff') or fn in index_files:
                os.unlink(join(patches_dir, fn))

    update_patches(eggs_dir, patches_dir, verbose)
    update_index(eggs_dir, patches_dir)


def main():
    from optparse import OptionParser

    p = OptionParser(
        usage="usage: %prog [options] [DIRECTORY]",
        description="updates egg patches, for a given egg repository.  "
                    "DIRECTORY defaults to CWD")

    p.add_option('-f', "--force", action="store_true")
    p.add_option('-v', "--verbose", action="store_true")

    opts, args = p.parse_args()

    if len(args) == 0:
        dir_path = os.getcwd()
    elif len(args) == 1:
        dir_path = abspath(args[0])
    else:
        p.error("too many arguments")

    update(dir_path, opts.force, opts.verbose)


if __name__ == '__main__':
    main()



import re
import os
import sys
import subprocess
import shutil
import zipfile
from os.path import dirname, expanduser, isdir, isfile, join
from optparse import OptionParser


verbose = False
eggs_dir = local_repo = prefix = pkgs_dir = python_exe = None


def unzip(zip_path, dir_path):
    """
    unpack the zip file into dir_path, creating directories as required
    """
    print "Unzipping: %r" % zip_path
    print "     into: %r" % dir_path
    z = zipfile.ZipFile(zip_path)
    for name in z.namelist():
        if name.endswith('/') or name.startswith('.unused'):
            continue
        path = join(dir_path, *name.split('/'))
        dpath = dirname(path)
        if not isdir(dpath):
            os.makedirs(dpath)
        fo = open(path, 'wb')
        fo.write(z.read(name))
        fo.close()
    z.close()


def parse_dists_file(path):
    pat = re.compile(r'(([+-]{2})\s*)?(\w\S+\.egg)$')
    for line in open(path):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        m = pat.match(line)
        if m is None:
            sys.exit('Error: invalid line in dists.txt: %r' % line)
        yield m.group(3)


def registry_egg(egg):
    return join(pkgs_dir, egg[:-4], 'EGG-INFO', 'registry.txt')


def cp_to_repo(egg, force=False):
    path = join(local_repo, egg)
    if not isfile(path) or force:
        shutil.copyfile(join(eggs_dir, egg), path)
    return path


def bootstrap_enstaller(egg):
    assert egg.startswith('enstaller-')
    code = ("import sys;"
            "sys.path.insert(0, %r);"
            "from egginst.bootstrap import main;"
            "main()" % cp_to_repo(egg))
    subprocess.check_call([python_exe, '-c', code])


def update_eggs(eggs):
    if not isdir(local_repo):
        os.makedirs(local_repo)

    if not isfile(python_exe):
        unzip(cp_to_repo(eggs[0]), prefix)

    if not isfile(registry_egg(eggs[1])):
        bootstrap_enstaller(eggs[1])

    if sys.platform == 'win32':
        egginst_py = join(prefix, r'Scripts\egginst-script.py')
    else:
        egginst_py = join(prefix, 'bin/egginst')

    for egg in eggs[1:]:
        if isfile(registry_egg(egg)):
            continue
        subprocess.check_call([python_exe, egginst_py, '--hook',
                               '--pkgs-dir', pkgs_dir,
                               cp_to_repo(egg)])


def main():
    if sys.platform == 'win32' and not sys.executable.endswith('python.exe'):
        default_eggs_dir = dirname(sys.executable)
    else:
        default_eggs_dir = os.getcwd()


    p = OptionParser(usage="usage: %prog [options]",
                     description=__doc__)

    p.add_option("--root",
                 action="store",
                 default=join(expanduser('~'), 'canopy'),
                 help="install location, defaults to %default",
                 metavar='PATH')

    p.add_option("--eggs-dir",
                 action="store",
                 default=default_eggs_dir,
                 help="directory with eggs to install (and dists.txt), "
                      "defaults to %default",
                 metavar='PATH')

    p.add_option('-v', "--verbose", action="store_true")

    opts, args = p.parse_args()

    if args:
        p.error('no argument expected, try -h')

    global verbose, eggs_dir, local_repo, prefix, pkgs_dir, python_exe

    eggs_dir = opts.eggs_dir
    verbose = opts.verbose

    if verbose:
        print "eggs_dir = %r" % eggs_dir

    eggs = list(parse_dists_file(join(eggs_dir, 'dists.txt')))

    if verbose:
        for egg in eggs:
            print "\t" + egg

    local_repo = join(opts.root, 'repo')
    prefix = join(opts.root, eggs[0][:-4])
    pkgs_dir = join(prefix, 'pkgs')
    if sys.platform == 'win32':
        python_exe = join(prefix, 'python.exe')
    else:
        python_exe = join(prefix, 'bin', 'python')

    if verbose:
        print "local_repo = %r" % local_repo
        print "prefix = %r" % prefix
        print "python_exe = %r" % python_exe

    update_eggs(eggs)

    entry_py = join(dirname(registry_egg(eggs[-1])), 'app_entry.py')
    return subprocess.call([python_exe, entry_py])


if __name__ == '__main__':
    sys.exit(main())



from cStringIO import StringIO
import hashlib
import json
import os
import os.path
import tempfile
import urllib2
from unittest import TestCase

from enstaller.store.cached import CachedHandler


class CacheTest(TestCase):
    def setUp(self):
        self.cache_dir = tempfile.mkdtemp()
        self.metadata_path = os.path.join(self.cache_dir, 'index_cache',
                                          'metadata.json')
        self.index_path = os.path.join(self.cache_dir, 'index_cache',
                                       'index.json')
        self.cache_handler = CachedHandler(self.cache_dir)

    def tearDown(self):
        for filename in os.listdir(self.cache_dir):
            try: os.remove(filename)
            except: pass
        try: os.rmdir(self.cache_dir)
        except: pass

    def _write_cache(self, data, metadata=None):
        os.mkdir(os.path.join(self.cache_dir, 'index_cache'))
        if data:
            open(self.index_path, 'wb').write(data)
        metadata = metadata or {'etag': 'test-etag',
                                'md5': hashlib.md5(data).hexdigest()}
        json.dump(metadata, open(self.metadata_path, 'wb'))

    def test_valid_cache(self):
        self._write_cache('foobly')
        self.assertTrue(self.cache_handler.cache_is_valid())

    def test_invalid_metadata(self):
        self._write_cache('foobly', {'etag': 'whatever', 'md5': 'notright'})
        self.assertFalse(self.cache_handler.cache_is_valid())

    def test_request_valid_cache(self):
        self._write_cache('foobly')
        req = urllib2.Request('http://whatever.com/index.json')
        req = self.cache_handler.http_request(req)
        self.assertEquals(req.headers['If-None-Match'], 'test-etag')

    def test_request_invalid_cache(self):
        """ No If-None-Match header should be added if there is no cached index.
        """
        self._write_cache(None, {'etag': 'whatever', 'md5': 'nomatter'})
        req = urllib2.Request('http://whatever.com/index.json')
        req = self.cache_handler.http_request(req)
        self.assertEquals(req.headers.get('If-None-Match'), None)

    def test_request_no_cache(self):
        req = urllib2.Request('http://whatever.com/index.json')
        req = self.cache_handler.http_request(req)
        self.assertEquals(req.headers.get('If-None-Match'), None)

    def test_write_cache(self):
        req = urllib2.Request('http://foo.com/index.json')
        res = urllib2.addinfourl(StringIO('hello'), {'Etag': 'returned-etag'}, 'http://redirected.com/index-foo.json')
        res.code = 200
        res.msg = 'OK'
        self.cache_handler.http_response(req, res)
        self.assertTrue(os.path.exists(self.index_path))
        self.assertTrue(os.path.exists(self.metadata_path))
        metadata = json.load(open(self.metadata_path, 'rb'))
        self.assertEqual(metadata['etag'], 'returned-etag')
        self.assertEqual(hashlib.md5(open(self.index_path, 'rb').read()).hexdigest(), metadata['md5'])

    def test_valid_304(self):
        self._write_cache('hello again')
        req = urllib2.Request('http://foo.com/index.json')
        res = self.cache_handler.http_error_304(req, StringIO(''), '304', 'Not Modified', {'Etag': 'test-etag'})
        self.assertEqual(res.read(), 'hello again')
        self.assertEqual(res.headers['Etag'], 'test-etag')

    def test_missing_304(self):
        """ On a 304 with an invalid cache, the cache should be cleared and a new request done. """

        class FakeParent(object):
            def open(self, url):
                return urllib2.addinfourl(StringIO('goodbye'), {}, 'http://whoop.com')

        self._write_cache('hello again')
        req = urllib2.Request('http://foo.com/index.json')
        self.cache_handler.parent = FakeParent()
        res = self.cache_handler.http_error_304(req, StringIO(''), '304', 'Not Modified', {'Etag': 'other-etag'})
        self.assertEqual(res.read(), 'goodbye')
        self.assertEqual(res.headers.get('Etag'), None)

    def test_no_etag(self):
        """ No Etag header == no cache written """
        req = urllib2.Request('http://foo.com/index.json')
        res = urllib2.addinfourl(StringIO('foo'), {}, 'http://foo.com/index.json')
        self.cache_handler.http_response(req, res)
        self.assertFalse(os.path.exists(self.metadata_path))
        self.assertFalse(os.path.exists(self.index_path))

    def test_ignore_non_index_etag(self):
        """ Etags are ignored for non-index requests """
        req = urllib2.Request('http://foo.com/')
        res = urllib2.addinfourl(StringIO('foo'), {'Etag': 'whatevs'}, 'http://foo.com/')
        self.cache_handler.http_response(req, res)
        self.assertFalse(os.path.exists(self.metadata_path))
        self.assertFalse(os.path.exists(self.index_path))

    def test_no_write_cache_on_304(self):
        """ Only write cache on 200 responses """
        req = urllib2.Request('http://foo.com/index.json')
        res = urllib2.addinfourl(StringIO('foo'), {'Etag': 'whatevs'}, 'http://foo.com/index.json')
        res.code = 304
        res.msg = 'Not Modified'
        self.cache_handler.http_response(req, res)
        self.assertFalse(os.path.exists(self.metadata_path))
        self.assertFalse(os.path.exists(self.index_path))




import sys
import hashlib
from os.path import abspath, expanduser, getmtime, getsize, isdir

from verlib import NormalizedVersion, IrrationalVersionError


PY_VER = '%i.%i' % sys.version_info[:2]


def abs_expanduser(path):
    return abspath(expanduser(path))


def canonical(s):
    """
    return the canonical representations of a project name
    DON'T USE THIS IN NEW CODE (ONLY (STILL) HERE FOR HISTORICAL REASONS)
    """
    # eventually (once Python 2.6 repo eggs are no longer supported), this
    # function should only return s.lower()
    s = s.lower()
    s = s.replace('-', '_')
    if s == 'tables':
        s = 'pytables'
    return s


def comparable_version(version):
    """
    Given a version string (e.g. '1.3.0.dev234'), return an object which
    allows correct comparison. For example:
        comparable_version('1.3.10') > comparable_version('1.3.8')  # True
    whereas:
        '1.3.10' > '1.3.8'  # False
    """
    try:
        # This hack makes it possible to use 'rc' in the version, where
        # 'rc' must be followed by a single digit.
        ver = version.replace('rc', '.dev99999')
        return NormalizedVersion(ver)
    except IrrationalVersionError:
        # If obtaining the RationalVersion object fails (for example for
        # the version '2009j'), simply return the string, such that
        # a string comparison can be made.
        return version


def md5_file(path):
    """
    Returns the md5sum of the file (located at `path`) as a hexadecimal
    string of length 32.
    """
    fi = open(path, 'rb')
    h = hashlib.new('md5')
    while True:
        chunk = fi.read(65536)
        if not chunk:
            break
        h.update(chunk)
    fi.close()
    return h.hexdigest()


def info_file(path):
    return dict(size=getsize(path),
                mtime=getmtime(path),
                md5=md5_file(path))


def cleanup_url(url):
    """
    Ensure a given repo string, i.e. a string specifying a repository,
    is valid and return a cleaned up version of the string.
    """
    if url.startswith(('http://', 'https://')):
        if not url.endswith('/'):
            url += '/'

    elif url.startswith('file://'):
        dir_path = url[7:]
        if dir_path.startswith('/'):
            # Unix filename
            if not url.endswith('/'):
                url += '/'
        else:
            # Windows filename
            if not url.endswith('\\'):
                url += '\\'

    elif isdir(abs_expanduser(url)):
        return cleanup_url('file://' + abs_expanduser(url))

    else:
        raise Exception("Invalid URL or non-existing file: %r" % url)

    return url


def fill_url(url):
    import plat

    url = url.replace('{ARCH}', plat.arch)
    url = url.replace('{SUBDIR}', plat.subdir)
    return cleanup_url(url)



"""
See:
www.surlyjake.com/2011/03/create-a-single-standalone-exe-from-a-python-program/


1.  grab pyinstaller 1.5rc (1.4 doesn't work with python 2.7).
    extract the zip file anywhere.

2. change directories to the pyinstaller folder you just created.

3. Before you create your first executable, you will have to run this once.

      python configure.py

4. Now, pyinstall needs to scan through your program and create what they
   call a spec file.

      python makespec.py --onefile path\to\program\program.py

5. Now, run this command to generate the executable.

      python build.py program\program.spec
"""
import os
import sys
import shutil
from subprocess import check_call
from os.path import abspath, isdir, join


pyinst_dir = r"C:\tmp\pyinstaller-1.5.1"


def main():
    subdir = join(pyinst_dir, "launch")
    if isdir(subdir):
        shutil.rmtree(subdir)

    check_call([sys.executable, "Makespec.py", "--onefile",
                abspath("launch.py")],
                cwd=pyinst_dir)

    check_call([sys.executable, "Build.py",
                join(subdir, "launch.spec")],
                cwd=pyinst_dir)

    shutil.copy(join(subdir, "dist", "launch.exe"), os.getcwd())


if __name__ == '__main__':
    main()



import shutil
import os
from os.path import dirname, isfile, join
from unittest import TestCase

from enstaller.main import Enstaller
from enstaller import config
from enstaller.indexed_repo.chain import Chain


class EnstallerTestCase(TestCase):
    def setUp(self):
        self.base_cfg = join(dirname(__file__), 'config',
                             '.enstaller4rc')
        test_cfg = self.base_cfg + '.copy'
        shutil.copy(self.base_cfg, test_cfg)
        config.home_config_path = test_cfg
        config.clear_cache()

        self.enst = Enstaller(Chain())

    def tearDown(self):
        config.clear_auth()

    def test_enstaller_userpass_upgrade(self):
        """ Username and password should be taken out of .enstaller4rc
        and stored in a keyring.
        """
        config.get_auth()
        config_contents = open(config.home_config_path).read()
        return # XXX
        self.assertTrue('EPD_auth' not in config_contents)
        username, password = config.get_auth()
        self.assertEqual(username, 'foo')
        self.assertEqual(password, 'bar')

        config.clear_auth()
        config_contents = open(config.home_config_path).read()
        self.assertTrue('EPD_username' not in config_contents)
        self.assertEqual(config.get_auth(), (None, None))

    def test_enstaller_userpass_new_config(self):
        """ Username and password should be stored in a keyring for
        brand-new config files, too
        """
        config.home_config_path = self.base_cfg + '.new'
        if isfile(config.home_config_path):
            os.unlink(config.home_config_path)
        config.clear_cache()

        config.change_auth('foo', 'bar')
        config_contents = open(config.home_config_path).read()
        return # XXX
        self.assertTrue('EPD_auth' not in config_contents)
        self.assertEqual(config.get_auth(), ('foo', 'bar'))

        config.clear_auth()
        self.assertEqual(config.get_auth(), (None, None))
        config.change_auth('foo', 'bar')
        self.assertEqual(config.get_auth(), ('foo', 'bar'))

    def test_enstaller_userpass_no_keyring(self):
        """ When the keyring module isn't available, username and
        password should be stored in the old method
        """
        keyring = config.keyring
        config.keyring = None
        try:
            config.get_auth()
            config_contents = open(config.home_config_path).read()
            self.assertTrue("EPD_auth = 'Zm9vOmJhcg=='" in config_contents)

            config.change_auth('bar', 'foo')
            config_contents = open(config.home_config_path).read()
            self.assertTrue("EPD_auth = 'YmFyOmZvbw=='" in config_contents)

            config.clear_auth()
            config_contents = open(config.home_config_path).read()
            self.assertTrue("EPD_auth" not in config_contents)
            self.assertEqual(config.get_auth(), (None, None))

            config.change_auth('foo', 'bar')
            self.assertEqual(config.get_auth(), ('foo', 'bar'))
        finally:
            config.keyring = keyring

    def test_enstaller_userpass_no_keyring_new_config(self):
        """ Username and password should be stored properly in a new
        config file, even with no keyring module
        """
        config.home_config_path = self.base_cfg + '.new'
        if isfile(config.home_config_path):
            os.unlink(config.home_config_path)
        config.clear_cache()

        keyring = config.keyring
        config.keyring = None
        try:
            config.change_auth('foo', 'bar')
            config_contents = open(config.home_config_path).read()
            self.assertTrue("EPD_auth = 'Zm9vOmJhcg=='" in config_contents)
            self.assertEqual(config.get_auth(), ('foo', 'bar'))
        finally:
            config.keyring = keyring




import re
import os
from collections import defaultdict
from os.path import basename, join, isdir, isfile, normpath, splitext



MODULE_EXTENSIONS = ('.pyd', '.so', '.py', '.pyw', '.pyc', 'pyo')
MODULE_EXTENSIONS_SET = set(MODULE_EXTENSIONS)


def stripped_lines(path):
    for line in open(path):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        yield line


def is_namespace(dir_path):
    modules = set()
    for fn in os.listdir(dir_path):
        path = join(dir_path, fn)
        if isfile(path):
            name, ext = splitext(basename(path))
            if ext in MODULE_EXTENSIONS_SET:
                modules.add(name)
    return modules == set(['__init__'])


id_pat = re.compile(r'[a-zA-Z_]\w*$')
def create_hooks_dir(dir_path, namespace=''):
    reg = {}
    modules = defaultdict(set)
    pth = []
    for fn in os.listdir(dir_path):
        name, ext = splitext(fn)
        if not id_pat.match(name):
            continue

        path = join(dir_path, fn)
        if isdir(path) and ext == '':
            reg[namespace + fn] = path
            if is_namespace(path):
                add_reg, dummy = create_hooks_dir(path, namespace + fn + '.')
                reg.update(add_reg)

        elif isfile(path):
            if ext in MODULE_EXTENSIONS_SET:
                modules[name].add(ext)

            elif ext == '.pth':
                for line in stripped_lines(path):
                    pth.append(normpath(join(dir_path, line)))

    for name, exts in modules.iteritems():
        if name == '__init__':
            continue
        for mext in MODULE_EXTENSIONS:
            if mext in exts:
                reg[namespace + name] = join(dir_path, name + mext)
                break

    return reg, pth


def create_file(egg):
    reg, pth = create_hooks_dir(egg.pkg_dir)
    def mk_rel(p):
        p = p.replace(egg.pkg_dir, '..').replace('\\', '/')
        if p == '..':
            return '../.'
        return p

    fo = open(egg.registry_txt, 'w')
    fo.write('# pkg: %s\n' % basename(egg.pkg_dir))
    for k, p in reg.iteritems():
        fo.write('%s  %s\n' % (k, mk_rel(p)))
    for p in pth:
        fo.write('-pth-  %s\n' % mk_rel(p))
    fo.close()


REGISTRY_CODE = """\
import sys
import imp
from os.path import abspath, dirname, join, splitext

EXT_INFO_MAP = {
    '.py': ('.py', 'U', imp.PY_SOURCE),
    '.pyw': ('.pyw', 'U', imp.PY_SOURCE),
    '.pyc': ('.pyc', 'rb', imp.PY_COMPILED),
    '.pyo': ('.pyo', 'rb', imp.PY_COMPILED),
    '.pyd': ('.pyd', 'rb', imp.C_EXTENSION),
    '.so': ('.so', 'rb', imp.C_EXTENSION),
    '': ('', '', imp.PKG_DIRECTORY),
}

class PackageRegistry(object):

    def __init__(self, registry={}):
        self.registry = registry
        self._path = None

    def find_module(self, fullname, path=None):
        try:
            self._path = self.registry[fullname]
            return self
        except KeyError:
            return None

    def load_module(self, fullname):
        mod = sys.modules.get(fullname)
        if mod:
            return mod
        assert self._path, "_path=%r" % self._path
        info = EXT_INFO_MAP[splitext(self._path)[-1]]
        if info[1]:
            f = open(self._path, info[1])
        else:
            f = None
        try:
            mod = imp.load_module(fullname, f, self._path, info)
        finally:
            if f is not None:
                f.close()
        return mod


def update_registry(paths):
    registry = {}
    for path in paths:
        for line in open(path):
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            k, v = line.split(None, 1)
            p = abspath(join(dirname(path), v))
            if k == '-pth-':
                if v not in sys.path:
                    sys.path.insert(0, p)
            else:
                registry[k] = p
    sys.meta_path.insert(0, PackageRegistry(registry))
"""



import re
import sys
import subprocess
from os.path import isfile, join

from enpkg import Enpkg


class Launch(Enpkg):

    def get_icon_path(self, egg):
        """ return the absolute path of the icon file

        the metadata 'app_icon_file' is the filename of the
        icon (which is always located in the meta-data directory
        """
        info = self.find(egg)
        if info and 'app_icon_file' in info:
            path = join(info['meta_dir'], info['app_icon_file'])
            return open(path, 'rb').read()
        return None

    def get_icon_data(self, egg):
        """ return the binary content of the icon file, either from the
        local file, or from the remote store when the egg is not installed
        """
        path = self.get_icon_path(egg)
        if path is None:
            # egg not installed, see if there is one in the store
            info = self.remote.get_metadata(egg)
            if 'app_icon_data' in info:
                return info['app_icon_data'].decode('base64')
        else:
            # egg is installed, so read the data from the file
            return open(path, 'rb').read()

    def launch_app(self, egg):
        info = self.find(egg)
        if 'app_cmd' in info:
            cmd = info['app_cmd']
        elif 'app_entry' in info:
            cmd = [sys.executable, join(info['meta_dir'], 'app_entry.py')]
        else:
            raise Exception("Don't know what to launch for egg: %r" % egg)
        if 'app_args' in info:
            cmd.extend(info['app_args'])
        if sys.platform == 'win32':
            DETACHED_PROCESS = 0x00000008
            subprocess.Popen(cmd, creationflags=DETACHED_FLAGS)
        else:
            subprocess.Popen(cmd)

    def registry_path_list(self, egg):
        if not self.hook:
            return []
        info = self.find(egg)
        pat = re.compile(r'([\w.]+)\s+([\w.]+)-(\d+)$')
        result = []
        for rs in ['%(name)s %(version)s-%(build)d' % info] + info['packages']:
            m = pat.match(rs)
            if not m:
                print "Warning: not a full requirement: %r" % rs
                continue
            d = self.find(m.expand(r'\1-\2-\3.egg'))
            if not d:
                print "Warning: cannot find install for: %r" % rs
                continue
            path = join(d['meta_dir'], 'registry.txt')
            if not isfile(path):
                print "Warning: no registry file:", path
                continue
            result.append(path)
        return result


if __name__ == '__main__':
    from enpkg import create_joined_store
    from plat import subdir

    urls = ['http://www.enthought.com/repo/.jpm/%s/' % subdir]

    remote = create_joined_store(urls)
    x = Launch(remote,
               #prefixes=['/home/ischnell/jpm/Python-2.7', sys.prefix],
               hook=1, verbose=1)

    fn = 'test-1.0-1.egg'

    x.install(fn)#, forceall=1)
    for k, info in x.query(app=True):
        print k
        if info.get('installed'):
            print '\t', x.get_icon(k)
    print x.get_icon(fn)
    x.launch_app(fn)



import json
import urlparse
import urllib2
from collections import defaultdict

from base import AbstractStore
from cached import CachedHandler
from compressed import CompressedHandler
from enstaller import config


class IndexedStore(AbstractStore):

    def connect(self, userpass=None):
        self.userpass = userpass  # tuple(username, password)

        self._index = self.get_index()

        #for k, v in self._index.iteritems():
        #    print k, v

        for info in self._index.itervalues():
            info['store_location'] = self.info().get('root')
            info.setdefault('type', 'egg')
            info.setdefault('python', '2.7')
            info.setdefault('packages', [])

        # maps names to keys
        self._groups = defaultdict(list)
        for key, info in self._index.iteritems():
            self._groups[info['name']].append(key)

    def get_index(self):
        fp = self.get_data('index.json')
        if fp is None:
            raise Exception("could not connect")
        return json.load(fp)

    def _location(self, key):
        rt = self.root.rstrip('/') + '/'
        if key.endswith('.zdiff'):
            return rt + 'patches/' + key
        return rt + key

    def get(self, key):
        return self.get_data(key), self.get_metadata(key)

    def get_metadata(self, key):
        return self._index[key]

    def exists(self, key):
        return key in self._index

    def query(self, **kwargs):
        for key in self.query_keys(**kwargs):
            yield key, self._index[key]

    def query_keys(self, **kwargs):
        name = kwargs.get('name')
        if name is None:
            for key, info in self._index.iteritems():
                if all(info.get(k) == v for k, v in kwargs.iteritems()):
                    yield key
        else:
            del kwargs['name']
            for key in self._groups[name]:
                info = self._index[key]
                if all(info.get(k) == v for k, v in kwargs.iteritems()):
                    yield key


class LocalIndexedStore(IndexedStore):

    def __init__(self, root_dir):
        self.root = root_dir

    def info(self):
        return dict(root=self.root)

    def get_data(self, key):
        try:
            return open(self._location(key), 'rb')
        except IOError as e:
            raise KeyError(str(e))


class RemoteHTTPIndexedStore(IndexedStore):

    def __init__(self, url):
        self.root = url
        self.opener = urllib2.build_opener(CompressedHandler,
                                           CachedHandler(config.get('local')),
                                           urllib2.HTTPHandler,
                                           urllib2.HTTPSHandler)

    def info(self):
        return dict(root=self.root)

    def get_index(self):
        fp = self.get_data('index.json?pypi=true')
        if fp is None:
            raise Exception("could not connect")
        return json.load(fp)

    def get_data(self, key):
        url = self._location(key)
        scheme, netloc, path, params, query, frag = urlparse.urlparse(url)
        auth, host = urllib2.splituser(netloc)
        if auth:
            auth = urllib2.unquote(auth)
        elif self.userpass:
            auth = ('%s:%s' % self.userpass)

        if auth:
            new_url = urlparse.urlunparse((scheme, host, path,
                                           params, query, frag))
            request = urllib2.Request(new_url)
            request.add_unredirected_header("Authorization",
                                "Basic " + auth.encode('base64').strip())
        else:
            request = urllib2.Request(url)
        request.add_header('User-Agent', 'enstaller')
        try:
            return self.opener.open(request)
        except urllib2.HTTPError as e:
            raise KeyError("%s: %s" % (e, url))
        except urllib2.URLError as e:
            raise Exception("Could not connect to %s" %(host,))



# Author: Ilan Schnell <ischnell@enthought.com>
"""\
enstaller is a managing tool for egginst-based installs, and the CLI is
called enpkg which calls out to egginst to do the actual install.
enpkg can access distributions from local and HTTP repositories.
"""
import os
import re
import sys
import site
import string
from argparse import ArgumentParser
from os.path import isfile, join

from egginst.utils import bin_dir_name, rel_site_packages
from enstaller import __version__
import config
from proxy.api import setup_proxy
from utils import abs_expanduser, fill_url

from eggcollect import EggCollection
from enpkg import Enpkg, EnpkgError, create_joined_store
from resolve import Req, comparable_info
from egg_meta import is_valid_eggname, split_eggname


FMT = '%-20s %-20s %s'
VB_FMT = '%(version)s-%(build)d'


def env_option(prefixes):
    print "Prefixes:"
    for p in prefixes:
        print '    %s%s' % (p, ['', ' (sys)'][p == sys.prefix])
    print

    cmd = ('export', 'set')[sys.platform == 'win32']
    print "%s PATH=%s" % (cmd, os.pathsep.join(
                                 join(p, bin_dir_name) for p in prefixes))
    if len(prefixes) > 1:
        print "%s PYTHONPATH=%s" % (cmd, os.pathsep.join(
                            join(p, rel_site_packages) for p in prefixes))

    if sys.platform != 'win32':
        if sys.platform == 'darwin':
            name = 'DYLD_LIBRARY_PATH'
        else:
            name = 'LD_LIBRARY_PATH'
        print "%s %s=%s" % (cmd, name, os.pathsep.join(
                                 join(p, 'lib') for p in prefixes))


def disp_store_info(info):
    sl = info.get('store_location')
    if not sl:
        return '-'
    for rm in 'http://', 'https://', 'www', '.enthought.com', '/repo/':
        sl = sl.replace(rm, '')
    return sl.replace('/eggs/', ' ').strip('/')


def name_egg(egg):
    return split_eggname(egg)[0]


def print_install_time(enpkg, name):
    for key, info in enpkg.ec.query(name=name):
        print '%s was installed on: %s' % (key, info['ctime'])


def info_option(enpkg, name):
    name = name.lower()
    print 'Package:', name
    versions = []
    for info in enpkg.info_list_name(name):
        versions.append(VB_FMT % info)
    print 'Available version: %s' % (', '.join(versions) or None)
    if versions:
        reqs = set(r for r in info['packages'])
        print "Requirements: %s" % (', '.join(sorted(reqs)) or None)
    print_install_time(enpkg, name)


def print_installed(prefix, hook=False, pat=None):
    print FMT % ('Name', 'Version', 'Store')
    print 60 * '='
    ec = EggCollection(prefix, hook)
    for egg, info in ec.query():
        if pat and not pat.search(info['name']):
            continue
        print FMT % (name_egg(egg), VB_FMT % info, disp_store_info(info))


def list_option(prefixes, hook=False, pat=None):
    for prefix in reversed(prefixes):
        print "prefix:", prefix
        print_installed(prefix, hook, pat)
        print


def parse_list(fn):
    pat = re.compile(r'([\w.]+)\s+([\w.]+-\d+)')
    res = set()
    for line in open(fn):
        line = line.strip()
        m = pat.match(line)
        if m:
            res.add(m.expand(r'\1-\2.egg'))
            continue
        if is_valid_eggname(line):
            res.add(line)
    return res


def imports_option(enpkg, pat=None):
    print FMT % ('Name', 'Version', 'Location')
    print 60 * "="

    names = set(info['name'] for _, info in enpkg.query_installed())
    for name in sorted(names, key=string.lower):
        if pat and not pat.search(name):
            continue
        for c in reversed(enpkg.ec.collections):
            index = dict(c.query(name=name))
            if index:
                info = index.values()[0]
                loc = 'sys' if c.prefix == sys.prefix else 'user'
        print FMT % (name, VB_FMT % info, loc)


def search(enpkg, pat=None):
    """
    print the packages which are available in the (remote) KVS
    """
    print FMT % ('Name', '  Versions', 'Note')
    print 60 * '-'

    names = {}
    for key, info in enpkg.query_remote():
        names[info['name']] = name_egg(key)

    installed = {}
    for key, info in enpkg.query_installed():
        installed[info['name']] = VB_FMT % info

    for name in sorted(names, key=string.lower):
        if pat and not pat.search(name):
            continue
        disp_name = names[name]
        installed_version = installed.get(name)
        for info in enpkg.info_list_name(name):
            version = VB_FMT % info
            disp_ver = (('* ' if installed_version == version else '  ') +
                        version)
            print FMT % (disp_name, disp_ver,
                   '' if info.get('available', True) else 'not subscribed to')
            disp_name = ''


def whats_new(enpkg):
    print FMT % ('Name', 'installed', 'available')
    print 60 * "="

    something_new = False
    for key, info in enpkg.query_installed():
        av_infos = enpkg.info_list_name(info['name'])
        if len(av_infos) == 0:
            continue
        av_info = av_infos[-1]
        if comparable_info(av_info) > comparable_info(info):
            print FMT % (name_egg(key), VB_FMT % info, VB_FMT % av_info)
            something_new = True

    if not something_new:
        print "no new version of any installed package is available"


def add_url(url, verbose):
    url = fill_url(url)
    if url in config.get('IndexedRepos'):
        print "Already configured:", url
        return
    config.prepend_url(url)


def install_req(enpkg, req, opts):
    try:
        actions = enpkg.install_actions(
                req,
                mode='root' if opts.no_deps else 'recur',
                force=opts.force, forceall=opts.forceall)
        enpkg.execute(actions)
    except EnpkgError, e:
        print e.message
        info_list = enpkg.info_list_name(req.name)
        if info_list:
            print "Versions for package %r are: %s" % (
                req.name,
                ', '.join(sorted(set(i['version'] for i in info_list))))
            if any(not i.get('available', True) for i in info_list):
                print "No subscription for %r" % req.name
        sys.exit(1)

    if len(actions) == 0:
        print "No update necessary, %r is up-to-date." % req.name
        print_install_time(enpkg, req.name)


def main():
    try:
        user_base = site.USER_BASE
    except AttributeError:
        user_base = abs_expanduser('~/.local')

    p = ArgumentParser(description=__doc__)
    p.add_argument('cnames', metavar='NAME', nargs='*',
                   help='package(s) to work on')
    p.add_argument("--add-url", metavar='URL',
                   help="add a repository URL to the configuration file")
    p.add_argument("--config", action="store_true",
                   help="display the configuration and exit")
    p.add_argument('-f', "--force", action="store_true",
                   help="force install the main package "
                        "(not it's dependencies, see --forceall)")
    p.add_argument("--forceall", action="store_true",
                   help="force install of all packages "
                        "(i.e. including dependencies)")
    p.add_argument("--hook", action="store_true",
                   help="don't install into site-packages (experimental)")
    p.add_argument("--imports", action="store_true",
                   help="show which packages can be imported")
    p.add_argument('-i', "--info", action="store_true",
                   help="show information about a package")
    p.add_argument("--log", action="store_true", help="print revision log")
    p.add_argument('-l', "--list", action="store_true",
                   help="list the packages currently installed on the system")
    p.add_argument('-n', "--dry-run", action="store_true",
               help="show what would have been downloaded/removed/installed")
    p.add_argument('-N', "--no-deps", action="store_true",
                   help="neither download nor install dependencies")
    p.add_argument("--env", action="store_true",
                   help="based on the configuration, display how to set the "
                        "some environment variables")
    p.add_argument("--prefix", metavar='PATH',
                   help="install prefix (disregarding of any settings in "
                        "the config file)")
    p.add_argument("--proxy", metavar='URL', help="use a proxy for downloads")
    p.add_argument("--remove", action="store_true", help="remove a package")
    p.add_argument("--revert", metavar="REV",
                   help="revert to a previous set of packages")
    p.add_argument('-s', "--search", action="store_true",
                   help="search the index in the repo of packages "
                        "and display versions available.")
    p.add_argument("--sys-config", action="store_true",
                   help="use <sys.prefix>/.enstaller4rc (even when "
                        "~/.enstaller4rc exists)")
    p.add_argument("--sys-prefix", action="store_true",
                   help="use sys.prefix as the install prefix")
    p.add_argument("--user", action="store_true",
               help="install into user prefix, i.e. --prefix=%r" % user_base)
    p.add_argument("--userpass", action="store_true",
                   help="change EPD authentication in configuration file")
    p.add_argument('-v', "--verbose", action="store_true")
    p.add_argument('--version', action="version",
                   version='enstaller version: ' + __version__)
    p.add_argument("--whats-new", action="store_true",
                   help="display to which installed packages updates are "
                        "available")
    args = p.parse_args()

    if len(args.cnames) > 0 and (args.config or args.env or args.userpass or
                                 args.revert or args.log or args.whats_new):
        p.error("Option takes no arguments")

    if args.user:
        args.prefix = user_base

    if args.prefix and args.sys_prefix:
        p.error("Options --prefix and --sys-prefix exclude each ohter")

    if args.force and args.forceall:
        p.error("Options --force and --forceall exclude each ohter")

    pat = None
    if (args.list or args.search) and args.cnames:
        pat = re.compile(args.cnames[0], re.I)

    # make prefix
    if args.sys_prefix:
        prefix = sys.prefix
    elif args.prefix:
        prefix = args.prefix
    else:
        prefix = config.get('prefix', sys.prefix)

    # now make prefixes
    if prefix == sys.prefix:
        prefixes = [sys.prefix]
    else:
        prefixes = [prefix, sys.prefix]

    if args.verbose:
        print "Prefixes:"
        for p in prefixes:
            print '    %s%s' % (p, ['', ' (sys)'][p == sys.prefix])
        print

    if args.env:                                  # --env
        env_option(prefixes)
        return

    if args.log:                                  # --log
        if args.hook:
            raise NotImplementedError
        from history import History
        h = History(prefix)
        h.update()
        h.print_log()
        return

    if args.sys_config:                           # --sys-config
        config.get_path = lambda: config.system_config_path

    if args.config:                               # --config
        config.print_config()
        return

    if args.list:                                 # --list
        list_option(prefixes, args.hook, pat)
        return

    if args.proxy:                                # --proxy
        setup_proxy(args.proxy)
    elif config.get('proxy'):
        setup_proxy(config.get('proxy'))
    else:
        setup_proxy()

    if 0: # for testing event manager only
        from encore.events.api import EventManager
        from encore.terminal.api import ProgressDisplay
        evt_mgr = EventManager()
        display = ProgressDisplay(evt_mgr)
    else:
        evt_mgr = None

    if config.get('use_webservice'):
        remote = None # Enpkg will create the default
    else:
        urls = [fill_url(u) for u in config.get('IndexedRepos')]
        remote = create_joined_store(urls)

    enpkg = Enpkg(remote, prefixes=prefixes, hook=args.hook,
                  evt_mgr=evt_mgr, verbose=args.verbose)

    if args.userpass:                             # --userpass
        auth = username, password = config.input_auth()
        if remote is not None:
            try:
                print 'Verifying username and password...'
                remote.connect(auth)
            except KeyError as e:
                print 'Invalid Username or Password'
            except Exception as e:
                print e.message
            else:
                config.change_auth(username, password)
        else:
            config.change_auth(username, password)
        return

    if args.dry_run:
        def print_actions(actions):
            for item in actions:
                print '%-8s %s' % item
        enpkg.execute = print_actions

    if args.imports:                              # --imports
        assert not args.hook
        imports_option(enpkg, pat)
        return

    if args.add_url:                              # --add-url
        add_url(args.add_url, args.verbose)
        return

    if args.revert:                               # --revert
        if isfile(args.revert):
            arg = parse_list(args.revert)
        else:
            arg = args.revert
        try:
            actions = enpkg.revert_actions(arg)
            if not actions:
                print "Nothing to do"
                return
            enpkg.execute(actions)
        except EnpkgError as e:
            print e.message
        return

    if args.search:                               # --search
        search(enpkg, pat)
        return

    if args.info:                                 # --info
        if len(args.cnames) != 1:
            p.error("Option requires one argument (name of package)")
        info_option(enpkg, args.cnames[0])
        return

    if args.whats_new:                            # --whats-new
        whats_new(enpkg)
        return

    if len(args.cnames) == 0:
        p.error("Requirement(s) missing")
    elif len(args.cnames) == 2:
        pat = re.compile(r'\d+\.\d+')
        if pat.match(args.cnames[1]):
            args.cnames = ['-'.join(args.cnames)]

    reqs = []
    for arg in args.cnames:
        if '-' in arg:
            name, version = arg.split('-', 1)
            reqs.append(Req(name + ' ' + version))
        else:
            reqs.append(Req(arg))

    if args.verbose:
        print "Requirements:"
        for req in reqs:
            print '    %r' % req
        print

    print "prefix:", prefix

    for req in reqs:
        if args.remove:                               # --remove
            try:
                enpkg.execute(enpkg.remove_actions(req))
            except EnpkgError as e:
                print e.message
        else:
            install_req(enpkg, req, args)             # install (default)


if __name__ == '__main__':
    main()



# unused
import os
import json
from os.path import isfile, join, getmtime, getsize

from base import AbstractStore



class CacheStore(AbstractStore):

    def __init__(self, remote, local):
        self.remote = remote
        self.local = local

    def connect(self, auth=None):
        pass

    def _copy_if_necessary(self, key):
        if not self.local.exists(key):
            self.local.set(key, self.remote.get(key))

    def get(self, key):
        self._copy_if_necessary(key)
        return self.local.get(key)

    def get_data(self, key):
        self._copy_if_necessary(key)
        return self.local.get_data(key)

    def get_metadata(self, key):
        #self._copy_if_necessary(key)
        return self.remote.get_metadata(key)

    def exists(self, key):
        return self.remote.exists(key)

    def query(self, **kwargs):
        return self.remote.query(**kwargs)

    def query_keys(self, **kwargs):
        return self.remote.query_keys(**kwargs)



import unittest
from os.path import dirname, join

from enstaller.history import History, find_revision


PATH = join(dirname(__file__), 'history')


class TestHistory(unittest.TestCase):

    def setUp(self):
        self.history = History('<dummy prefix>')
        self.history._log_path = PATH

    def test_find_revision(self):
        times = [
            '2011-08-01 21:17:22 CDT',
            '2011-08-01 22:38:37 CDT',
            '2011-08-01 23:05:07 CDT',
        ]
        for dt, res in [
            ('2011-08-01 21:17:21 CDT', 0),
            ('2011-08-01 21:17:22 CDT', 0),
            ('2011-08-01 21:17:23 CDT', 0),
            ('2011-08-01 22:38:36 CDT', 0),
            ('2011-08-01 22:38:37 CDT', 1),
            ('2011-08-01 22:38:38 CDT', 1),
            ('2011-08-01 23:05:06 CDT', 1),
            ('2011-08-01 23:05:07 CDT', 2),
            ('2011-08-01 23:05:08 CDT', 2),
            ]:
            self.assertEqual(find_revision(times, dt), res)

    def test_get_state(self):
        self.assertEqual(self.history.get_state(0),
                         set(['appinst-2.1.0-1.egg',
                              'basemap-1.0.1-1.egg',
                              'biopython-1.57-2.egg']))
        self.assertEqual(self.history.get_state(),
                         set(['basemap-1.0.2-1.egg',
                              'biopython-1.57-2.egg',
                              'numpy-1.7.0-1.egg']))


if __name__ == '__main__':
    unittest.main()



# Copyright by Enthought, Inc.
# Author: Ilan Schnell <ischnell@enthought.com>

import re
import os
import sys
import platform
from os.path import isfile, join

from enstaller import __version__
from utils import PY_VER, abs_expanduser, fill_url


try:
    import keyring
    import keyring.backend
    # don't use keyring backends that require console input or just do
    # more or less the same thing we're already doing
    keyring.backend._all_keyring = [keyring.backend.OSXKeychain(),
                                    keyring.backend.GnomeKeyring(),
                                    keyring.backend.KDEKWallet(),
                                    keyring.backend.Win32CryptoKeyring(),
                                    keyring.backend.Win32CryptoRegistry(),
                                    keyring.backend.WinVaultKeyring()]
    keyring.core.init_backend()
    if keyring.get_keyring().supported() < 0:
        keyring = None
except (ImportError, KeyError):
    # the KeyError happens on Windows when the environment variable
    # 'USERPROFILE' is not set
    keyring = None

KEYRING_SERVICE_NAME = 'Enthought.com'

config_fn = ".enstaller4rc"
home_config_path = abs_expanduser("~/" + config_fn)
system_config_path = join(sys.prefix, config_fn)

default = dict(
    prefix=sys.prefix,
    proxy=None,
    noapp=False,
    local=join(sys.prefix, 'LOCAL-REPO'),
    EPD_auth=None,
    EPD_userpass=None,
    use_webservice=True,
    IndexedRepos=[],
)


def get_path():
    """
    Return the absolute path to our config file.
    """
    if isfile(home_config_path):
        return home_config_path
    if isfile(system_config_path):
        return system_config_path
    return None


def input_auth():
    from getpass import getpass
    print """\
In order to access the EPD repository, please enter your
username and password, which you use to subscribe to EPD.
If you are not subscribed to EPD, just hit Return.
"""
    username = raw_input('Username: ').strip()
    if not username:
        return None, None
    return username, getpass('Password: ')


RC_TMPL = """\
# enstaller configuration file
# ============================
#
# This file contains the default package repositories, and configuration,
# used by enstaller %(version)s for the Python %(py_ver)s environment:
#
#   sys.prefix = %(sys_prefix)r
#
# This file was created by initially running the enpkg command.

%(auth_section)s

# use_webservice refers to using 'https://api.enthought.com/eggs/',
# the default is True, i.e. the webservice URL is used for fetching eggs.
# Uncommenting changes this behaviour, to use the explicit IndexedRepos
# listed below.
#use_webservice = False

# The enpkg command is searching for eggs in the list 'IndexedRepos'.
# When enpkg is searching for an egg, it tries to find it in the order
# of this list, and selects the first one that matches, ignoring
# repositories below.  Therefore the order of this list matters.
#
# For local repositories, the index file is optional.  Remember that on
# Windows systems the backslaches in the directory path need to escaped, e.g.:
# r'file://C:\\repository\\' or 'file://C:\\\\repository\\\\'
IndexedRepos = [
#  'https://www.enthought.com/repo/ets/eggs/{SUBDIR}/',
  'https://www.enthought.com/repo/epd/GPL-eggs/{SUBDIR}/',
  'https://www.enthought.com/repo/epd/eggs/{SUBDIR}/',
# The Enthought PyPI build mirror:
  'http://www.enthought.com/repo/pypi/eggs/{SUBDIR}/',
]

# Install prefix (enpkg --prefix and --sys-prefix options overwrite this).
# When this variable is not provided, it will default to the value of
# sys.prefix (within the current interpreter running enpkg)
#prefix = %(sys_prefix)r

# When running enpkg behind a firewall it might be necessary to use a proxy
# to access the repositories.  The URL for the proxy can be set here.
# Note that the enpkg --proxy option will overwrite this setting.
%(proxy_line)s

# Uncommenting the next line will disable application menu item install.
# This only effects the few packages which install menu items,
# which as IPython.
#noapp = True
"""

def write(username=None, password=None, proxy=None):
    """
    write the config file
    """
    # If user is 'root', then always create the config file in sys.prefix,
    # otherwise in the user's HOME directory.
    if sys.platform != 'win32' and os.getuid() == 0:
        path = system_config_path
    else:
        path = home_config_path

    if username is None and password is None:
        username, password = input_auth()
    if username and password:
        if keyring:
            keyring.set_password(KEYRING_SERVICE_NAME, username, password)
            authline = 'EPD_username = %r' % username
        else:
            auth = ('%s:%s' % (username, password)).encode('base64')
            authline = 'EPD_auth = %r' % auth.strip()
        auth_section = """
# The EPD subscriber authentication is required to access the EPD repository.
# To change this setting, use the 'enpkg --userpass' command which will ask
# you for your username and password.
%s
""" % authline
    else:
        auth_section = ''

    py_ver = PY_VER
    sys_prefix = sys.prefix
    version = __version__

    if proxy:
        proxy_line = 'proxy = %r' % proxy
    else:
        proxy_line = '#proxy = <proxy string>  # e.g. "123.0.1.2:8080"'

    fo = open(path, 'w')
    fo.write(RC_TMPL % locals())
    fo.close()
    print "Wrote configuration file:", path
    print 77 * '='
    clear_cache()


def get_auth():
    password = None
    old_auth = get('EPD_auth')
    if old_auth:
        username, password = old_auth.decode('base64').split(':')
        if not keyring:
            return username, password
        else:
            change_auth(username, password)
    username = get('EPD_username')
    if username and keyring:
        if hasattr(get_auth, 'password'):
            password = get_auth.password
        else:
            password = keyring.get_password(KEYRING_SERVICE_NAME, username)
            get_auth.password = password
    if username and password:
        return username, password
    else:
        return None, None


def clear_auth():
    username = get('EPD_username')
    if username and keyring:
        keyring.set_password(KEYRING_SERVICE_NAME, username, '')
    change_auth('', None)


def change_auth(username, password):
    # clear the cache so the next get_auth is correct
    clear_cache()

    path = get_path()
    if path is None:
        write(username, password)
        return
    fi = open(path)
    data = fi.read()
    fi.close()

    if username is None and password is None:
        return

    if username and password:
        if keyring:
            keyring.set_password(KEYRING_SERVICE_NAME, username, password)
            authline = 'EPD_username = %r' % username
        else:
            auth = ('%s:%s' % (username, password)).encode('base64')
            authline = 'EPD_auth = %r' % auth.strip()

    pat = re.compile(r'^(EPD_auth|EPD_username)\s*=.*$', re.M)

    if username == '':
        data = pat.sub('', data)
    elif pat.search(data):
        data = pat.sub(authline, data)
    else:
        lines = data.splitlines()
        lines.insert(10, authline)
        data = '\n'.join(lines) + '\n'
    fo = open(path, 'w')
    fo.write(data)
    fo.close()


def prepend_url(url):
    f = open(get_path(), 'r+')
    data = f.read()
    pat = re.compile(r'^IndexedRepos\s*=\s*\[\s*$', re.M)
    if not pat.search(data):
        sys.exit("Error: IndexedRepos section not found")
    data = pat.sub(r"IndexedRepos = [\n  '%s'," % url, data)
    f.seek(0)
    f.write(data)
    f.close()


def clear_cache():
    if hasattr(read, 'cache'):
        del read.cache


def read():
    """
    return the configuration from the config file as a dictionary,
    and fix some values and give defaults
    """
    if hasattr(read, 'cache'):
        return read.cache

    path = get_path()
    read.cache = default
    if path is None:
        return read.cache

    execfile(path, read.cache)
    for k in read.cache:
        v = read.cache[k]
        if k == 'IndexedRepos':
            read.cache[k] = [fill_url(url) for url in v]
        elif k in ('prefix', 'local'):
            read.cache[k] = abs_expanduser(v)
    return read.cache


def get(key, default=None):
    return read().get(key, default)


def print_config():
    print "Python version:", PY_VER
    print "enstaller version:", __version__
    print "sys.prefix:", sys.prefix
    print "platform:", platform.platform()
    print "architecture:", platform.architecture()[0]
    print "use_webservice:", get('use_webservice')
    print "config file:", get_path()
    print
    print "settings:"
    for k in 'prefix', 'local', 'noapp', 'proxy':
        print "    %s = %r" % (k, get(k))
    print "    IndexedRepos:", '(not used)' if get('use_webservice') else ''
    for repo in get('IndexedRepos'):
        print '        %r' % repo


if __name__ == '__main__':
    write()
    print_config()



"""
A .zdiff is a diff of two zip-files.
It is an (uncompressed) zip-file where each archive contains information
about how SRC needs to be modified to obtain DST.  A non-existing archive
means the SRC is unchanged.  The data in the archives always has to start
with one of the following:
  * BSDIFF4: the following data is a binary diff between SRC and DST
  * BZ: the new data of DST (bz2 compressed), SRC is ignored
  * RM: DST does not exist (it needs removed from SRC)

"""
import bz2
import json
import zipfile
from uuid import uuid4
from os.path import basename, getmtime, getsize

from utils import md5_file

import bsdiff4


def diff(src_path, dst_path, patch_path):
    x = zipfile.ZipFile(src_path)
    y = zipfile.ZipFile(dst_path)
    z = zipfile.ZipFile(patch_path, 'w', zipfile.ZIP_STORED)

    xnames = set(x.namelist())
    ynames = set(y.namelist())

    count = 0
    for name in xnames | ynames:
        xdata = x.read(name) if name in xnames else None
        ydata = y.read(name) if name in ynames else None
        if xdata == ydata:
            continue

        if ydata is not None:
            bz2_data = bz2.compress(ydata) # startswith BZ

        if xdata is not None and ydata is not None:
            diff_data = bsdiff4.diff(xdata, ydata)
            if len(diff_data) < len(bz2_data):
                zdata = diff_data # startswith BSDIFF4
            else:
                zdata = bz2_data
        elif xdata is not None and ydata is None:
            zdata = 'RM'
        elif ydata is not None and xdata is None:
            zdata = bz2_data
        else:
            raise Exception("Hmm, didn't expect to get here.")

        #print zdata[:2], name
        z.writestr(name, zdata)
        count += 1

    info = {}
    for path, pre in (src_path, 'src'), (dst_path, 'dst'):
        info.update({pre: basename(path),
                     pre + '_size': getsize(path),
                     pre + '_mtime': getmtime(path),
                     pre + '_md5': md5_file(path)})
    z.writestr('__zdiff_info__.json',
               json.dumps(info, indent=2, sort_keys=True))
    z.close()
    y.close()
    x.close()
    return count


def patch(src_path, dst_path, patch_path, evt_mgr=None, super_id=None):
    if evt_mgr:
        from encore.events.api import ProgressManager
    else:
        from egginst.console import ProgressManager

    x = zipfile.ZipFile(src_path)
    y = zipfile.ZipFile(dst_path, 'w', zipfile.ZIP_DEFLATED)
    z = zipfile.ZipFile(patch_path)

    xnames = x.namelist()
    znames = set(z.namelist())

    n = 0
    tot = len(xnames) + len(znames)

    progress = ProgressManager(
                evt_mgr, source=patch,
                operation_id=uuid4(),
                message="patching",
                steps=tot,
                # ---
                progress_type="patching", filename=basename(patch_path),
                disp_amount=str(tot),
                super_id=super_id)

    with progress:
        for name in xnames:
            if name not in znames:
                 y.writestr(x.getinfo(name), x.read(name))
            n += 1
            progress(step=n)

        for name in z.namelist():
            if name == '__zdiff_info__.json':
                continue
            zdata = z.read(name)
            if zdata.startswith('BSDIFF4'):
                ydata = bsdiff4.patch(x.read(name), zdata)
            elif zdata.startswith('BZ'):
                ydata = bz2.decompress(zdata)
            elif zdata.startswith('RM'):
                continue
            else:
                raise Exception("Hmm, didn't expect to get here: %r" % zdata)

            y.writestr(name, ydata)
            progress(step=n)

    z.close()
    y.close()
    x.close()


def info(patch_path):
    z = zipfile.ZipFile(patch_path)
    data = z.read('__zdiff_info__.json')
    z.close()
    return json.loads(data)


if __name__ == '__main__':
    import sys
    if len(sys.argv) == 3:
        src, dst = sys.argv[1:3]
    else:
        src, dst = 'nose-1.0.0-1.egg', 'nose-1.1.2-1.egg'

    pat = 'test.zdiff'
    diff(src, dst, pat)
    patch(src, dst + '2', pat)
    print info(pat)
    assert diff(dst, dst + '2', 'dummy.zdiff') == 0



import os
import sys
import imp
from os.path import isfile, join, splitext


EXT_INFO_MAP = {
    '.py': ('.py', 'U', imp.PY_SOURCE),
    '.pyw': ('.pyw', 'U', imp.PY_SOURCE),
    '.pyc': ('.pyc', 'rb', imp.PY_COMPILED),
    '.pyo': ('.pyo', 'rb', imp.PY_COMPILED),
    '.pyd': ('.pyd', 'rb', imp.C_EXTENSION),
    '.so': ('.so', 'rb', imp.C_EXTENSION),
    '': ('', '', imp.PKG_DIRECTORY),
}


class PackageRegistry(object):

    def __init__(self, registry={}):
        self.registry = registry
        self._path = None

    def find_module(self, fullname, path=None):
        """
        Find fullname in the registry if it is recorded directly or if any
        of the partial names are recorded.
        That is, if there is a registry entry for scipy.stats then
        scipy.stats.distributions will be found underneath the path tree
        where it is located.
        """
        # sys.stderr.write("# registry find_module: %s (path=%s)\n" %
        #                  (fullname, path))
        try:
            self._path = self.registry[fullname]
            return self
        except KeyError:
            return None

    def load_module(self, fullname):
        """
        uses imp.load_module
        """
        # print "Called load_module", fullname
        mod = sys.modules.get(fullname)
        if mod:
            return mod

        if sys.flags.verbose:
            sys.stderr.write("# registry load_module: %s\n" % fullname)
        assert self._path, "_path=%r" % self._path

        info = EXT_INFO_MAP[splitext(self._path)[-1]]
        if info[1]:
            f = open(self._path, info[1])
        else:
            f = None

        try:
            mod = imp.load_module(fullname, f, self._path, info)
        finally:
            if f is not None:
                f.close()

        return mod


def update(path):
    """
    updates the registry and sys.path given the path to a registry file

    The registry file should have one entry per package and additional
    entries for any individual files that don't fit under any package
    (essentially) anything that is top level.
    In addition, it may contain "-pth-" entries, which are simply appended
    to sys.path.
    """
    registry = {}
    for line in open(path):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        k, v = line.split(None, 1)
        if k == '-pth-':
            if v not in sys.path:
                sys.path.insert(0, v)
        else:
            registry[k] = v

    sys.meta_path.insert(0, PackageRegistry(registry))


def main():
    """
    called from main() in site.py
    """
    path = os.environ.get('EPDREGISTRY', join(sys.prefix, 'registry.txt'))
    if sys.flags.verbose:
        sys.stderr.write("# registry file: %s\n" % path)

    if isfile(path):
        update(path)
    elif sys.flags.verbose:
        sys.stderr.write("# WARNING: registry file does not exist\n")


if __name__ == '__main__':
    main()



import unittest
from collections import defaultdict
from os.path import abspath, dirname, join

from enstaller.store.indexed import IndexedStore
from enstaller.store.joined import JoinedStore

from enstaller import resolve
from enstaller.resolve import Resolve, Req
from enstaller.indexed_repo.metadata import parse_depend_index


this_dir = abspath(dirname(__file__))


class DummyStore(IndexedStore):

    def __init__(self, index_path, name=None):
        self.index_path = index_path
        self.name = name

    def connect(self, auth=None):
        index_data = open(self.index_path).read()
        self._index = parse_depend_index(index_data)
        for spec in self._index.itervalues():
            spec['name'] = spec['name'].lower()
            spec['type'] = 'egg'
            spec['repo_dispname'] = self.name
        self._groups = defaultdict(list)
        for key, info in self._index.iteritems():
            self._groups[info['name']].append(key)

    def get_data(self, key):
        pass


def eggs_rs(c, req_string):
    return c.install_sequence(Req(req_string))


class TestReq(unittest.TestCase):

    def test_init(self):
        for req_string, name, version, build, strictness in [
            ('',          None,  None,  None, 0),
            (' \t',       None,  None,  None, 0),
            ('foo',       'foo', None,  None, 1),
            (u'bar 1.9',  'bar', '1.9', None, 2),
            ('BAZ 1.8-2', 'baz', '1.8', 2,    3),
            ('qux 1.3-0', 'qux', '1.3', 0,    3),
            ]:
            r = Req(req_string)
            self.assertEqual(r.name, name)
            self.assertEqual(r.version, version)
            self.assertEqual(r.build, build)
            self.assertEqual(r.strictness, strictness)

    def test_as_dict(self):
        for req_string, d in [
            ('',          dict()),
            ('foo',       dict(name='foo')),
            ('bar 1.9',   dict(name='bar', version='1.9')),
            ('BAZ 1.8-2', dict(name='baz', version='1.8', build=2)),
            ]:
            r = Req(req_string)
            self.assertEqual(r.as_dict(), d)

    def test_misc_methods(self):
        for req_string in ['', 'foo', 'bar 1.2', 'baz 2.6.7-5']:
            r = Req(req_string)
            self.assertEqual(str(r), req_string)
            self.assertEqual(r, r)
            self.assertEqual(eval(repr(r)), r)

        self.assertNotEqual(Req('foo'), Req('bar'))
        self.assertNotEqual(Req('foo 1.4'), Req('foo 1.4-5'))

    def test_matches(self):
        spec = dict(name='foo_bar', version='2.4.1', build=3, python=None)
        for req_string, m in [
            ('', True),
            ('foo', False),
            ('Foo_BAR', True),
            ('foo_Bar 2.4.1', True),
            ('FOO_Bar 1.8.7', False),
            ('FOO_BAR 2.4.1-3', True),
            ('FOO_Bar 2.4.1-1', False),
            ]:
            self.assertEqual(Req(req_string).matches(spec), m, req_string)

    def test_matches_py(self):
        spec = dict(name='foo', version='2.4.1', build=3, python=None)
        for py in ['2.4', '2.5', '2.6', '3.1']:
            resolve.PY_VER = py
            self.assertEqual(Req('foo').matches(spec), True)

        spec25 = dict(spec)
        spec25.update(dict(python='2.5'))

        spec26 = dict(spec)
        spec26.update(dict(python='2.6'))

        resolve.PY_VER = '2.5'
        self.assertEqual(Req('foo').matches(spec25), True)
        self.assertEqual(Req('foo').matches(spec26), False)

        resolve.PY_VER = '2.6'
        self.assertEqual(Req('foo').matches(spec25), False)
        self.assertEqual(Req('foo').matches(spec26), True)


class TestChain0(unittest.TestCase):

    r = JoinedStore([
           DummyStore(join(this_dir, fn))
           for fn in ['index-add.txt', 'index-5.1.txt', 'index-5.0.txt']])
    r.connect()
    c = Resolve(r)

    def test_25(self):
        resolve.PY_VER = '2.5'
        self.assertEqual(eggs_rs(self.c, 'SciPy 0.8.0.dev5698'),
                         ['freetype-2.3.7-1.egg', 'libjpeg-7.0-1.egg',
                          'numpy-1.3.0-1.egg', 'PIL-1.1.6-4.egg',
                          'scipy-0.8.0.dev5698-1.egg'])

        self.assertEqual(eggs_rs(self.c, 'SciPy'),
                         ['numpy-1.3.0-1.egg', 'scipy-0.8.0-1.egg'])

        self.assertEqual(eggs_rs(self.c, 'epdcore'),
                         ['AppInst-2.0.4-1.egg', 'numpy-1.3.0-1.egg',
                          'scipy-0.8.0-1.egg', 'EPDCore-1.2.5-1.egg'])

    def test_26(self):
        resolve.PY_VER = '2.6'

        self.assertEqual(eggs_rs(self.c, 'SciPy'),
                         ['numpy-1.3.0-2.egg', 'scipy-0.8.0-2.egg'])

        self.assertEqual(eggs_rs(self.c, 'epdcore'),
                         ['numpy-1.3.0-2.egg', 'scipy-0.8.0-2.egg',
                          'EPDCore-2.0.0-1.egg'])

class TestChain1(unittest.TestCase):

    r = JoinedStore([
            DummyStore(join(this_dir, name, 'index-7.1.txt'), name)
            for name in ('epd', 'gpl')])
    r.connect()
    c = Resolve(r)

    def test_get_repo(self):
        for req_string, repo_name in [
            ('MySQL_python', 'gpl'),
            ('bitarray', 'epd'),
            ('foobar', None),
            ]:
            egg = self.c.get_egg(Req(req_string))
            if egg is not None:
                self.assertEqual(
                    self.r.get_metadata(egg).get('repo_dispname'),
                    repo_name)

    def test_get_dist(self):
        resolve.PY_VER = '2.7'
        for req_string, repo_name, egg in [
            ('MySQL_python',  'gpl', 'MySQL_python-1.2.3-2.egg'),
            ('numpy',         'epd', 'numpy-1.6.0-3.egg'),
            ('swig',          'epd', 'swig-1.3.40-2.egg'),
            ('swig 1.3.36',   'epd', 'swig-1.3.36-3.egg'),
            ('swig 1.3.40-1', 'epd', 'swig-1.3.40-1.egg'),
            ('swig 1.3.40-2', 'epd', 'swig-1.3.40-2.egg'),
            ('foobar', None, None),
            ]:
            self.assertEqual(self.c.get_egg(Req(req_string)), egg)
            if egg is not None:
                self.assertEqual(
                    self.r.get_metadata(egg).get('repo_dispname'),
                    repo_name)

    def test_reqs_dist(self):
        self.assertEqual(self.c.reqs_egg('FiPy-2.1-1.egg'),
                         set([Req('distribute'),
                              Req('scipy'),
                              Req('numpy'),
                              Req('pysparse 1.2.dev203')]))

    def test_root(self):
        self.assertEqual(self.c.install_sequence(Req('numpy 1.5.1'),
                                                 mode='root'),
                         ['numpy-1.5.1-2.egg'])

        self.assertEqual(self.c.install_sequence(Req('numpy 1.5.1-1'),
                                                 mode='root'),
                         ['numpy-1.5.1-1.egg'])

    def test_order1(self):
        self.assertEqual(self.c.install_sequence(Req('numpy')),
                         ['MKL-10.3-1.egg', 'numpy-1.6.0-3.egg'])

    def test_order2(self):
        self.assertEqual(self.c.install_sequence(Req('scipy')),
                         ['MKL-10.3-1.egg', 'numpy-1.5.1-2.egg',
                          'scipy-0.9.0-1.egg'])


class TestChain2(unittest.TestCase):

    r = JoinedStore([
            DummyStore(join(this_dir, name, 'index-7.1.txt'), name)
            for name in ('open', 'runner', 'epd')])
    r.connect()
    c = Resolve(r)

    def test_flat_recur1(self):
        d1 = self.c.install_sequence(Req('openepd'), mode='flat')
        d2 = self.c.install_sequence(Req('openepd'), mode='recur')
        self.assertEqual(d1, d2)
        d3 = self.c.install_sequence(Req('foo'), mode='recur')
        self.assertEqual(d2[:-1], d3[:-1])

    def test_flat_recur2(self):
        for rs in 'epd 7.0', 'epd 7.0-1', 'epd 7.0-2':
            d1 = self.c.install_sequence(Req(rs), mode='flat')
            d2 = self.c.install_sequence(Req(rs), mode='recur')
            self.assertEqual(d1, d2)

    def test_multiple_reqs(self):
        lst = self.c.install_sequence(Req('ets'))
        self.assert_('numpy-1.5.1-2.egg' in lst)
        self.assertEqual(
            self.r.get_metadata('numpy-1.5.1-2.egg').get('repo_dispname'),
            'epd')


if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod


class AbstractStore(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def connect(self, authentication=None):
        raise NotImplementedError

    def info(self):
        raise NotImplementedError

    @abstractmethod
    def get(self, key):
        raise NotImplementedError

    def set(self, key, value, buffer_size=1048576):
        raise NotImplementedError

    def delete(self, key):
        raise NotImplementedError

    @abstractmethod
    def get_data(self, key):
        raise NotImplementedError

    @abstractmethod
    def get_metadata(self, key, select=None):
        raise NotImplementedError

    def set_data(self, key, data):
        raise NotImplementedError

    def set_metadata(self, key, metadata):
        raise NotImplementedError

    def update_metadata(self, key, metadata):
        raise NotImplementedError

    @abstractmethod
    def exists(self, key):
        raise NotImplementedError

    @abstractmethod
    def query(self, select=None, **kwargs):
        raise NotImplementedError


    def query_keys(self, **kwargs):
        return self.query(**kwargs).keys()



import math
import os
import sys
import hashlib
from uuid import uuid4
from os.path import basename, isdir, isfile, join

from egginst.utils import human_bytes, rm_rf
from utils import md5_file


class FetchAPI(object):

    def __init__(self, remote, local_dir, evt_mgr=None):
        self.remote = remote
        self.local_dir = local_dir
        self.evt_mgr = evt_mgr
        self.verbose = False

    def path(self, fn):
        return join(self.local_dir, fn)

    def fetch(self, key):
        path = self.path(key)
        fi, info = self.remote.get(key)

        size = info['size']
        md5 = info.get('md5')

        if self.evt_mgr:
            from encore.events.api import ProgressManager
        else:
            from egginst.console import ProgressManager

        progress = ProgressManager(
                self.evt_mgr, source=self,
                operation_id=uuid4(),
                message="fetching",
                steps=size,
                # ---
                progress_type="fetching", filename=basename(path),
                disp_amount=human_bytes(size),
                super_id=getattr(self, 'super_id', None))

        n = 0
        h = hashlib.new('md5')
        if size < 256:
            buffsize = 1
        else:
            buffsize = 2 ** int(math.log(size / 256.0) / math.log(2.0) + 1)

        pp = path + '.part'
        if sys.platform == 'win32':
            rm_rf(pp)
        with progress:
            with open(pp, 'wb') as fo:
                while True:
                    chunk = fi.read(buffsize)
                    if not chunk:
                        break
                    fo.write(chunk)
                    if md5:
                        h.update(chunk)
                    n += len(chunk)
                    progress(step=n)
        fi.close()

        if md5 and h.hexdigest() != md5:
            raise ValueError("received data MD5 sums mismatch")

        if sys.platform == 'win32':
            rm_rf(path)
        os.rename(pp, path)

    def patch_egg(self, egg):
        """
        Try to create 'egg' by patching an already existing egg, returns
        True on success and False on failure, i.e. when either:
            - bsdiff4 is not installed
            - no patches can be applied because: (i) there are no relevant
              patches in the repo (ii) a source egg is missing
        """
        try:
            import enstaller.zdiff as zdiff
        except ImportError:
            if self.verbose:
                print "Warning: could not import bsdiff4, cannot patch"
            return False

        possible = []
        for patch_fn, info in self.remote.query(
                          type='patch',
                          name=egg.split('-')[0].lower(),
                          dst=egg):
            assert info['dst'] == egg
            src_path = self.path(info['src'])
            #print '%8d %s %s' % (info['size'], patch_fn, isfile(src_path))
            if isfile(src_path):
                possible.append((info['size'], patch_fn, info))

        if not possible:
            return False
        size, patch_fn, info = min(possible)

        self.fetch(patch_fn)
        zdiff.patch(self.path(info['src']), self.path(egg),
                    self.path(patch_fn), self.evt_mgr,
                    super_id=getattr(self, 'super_id', None))
        return True

    def fetch_egg(self, egg, force=False):
        """
        fetch an egg, i.e. copy or download the distribution into local dir
        force: force download or copy if MD5 mismatches
        """
        if not isdir(self.local_dir):
            os.makedirs(self.local_dir)
        info = self.remote.get_metadata(egg)
        path = self.path(egg)

        # if force is used, make sure the md5 is the expected, otherwise
        # merely see if the file exists
        if isfile(path):
            if force:
                if md5_file(path) == info.get('md5'):
                    if self.verbose:
                        print "Not refetching, %r MD5 match" % path
                    return
            else:
                if self.verbose:
                    print "Not forcing refetch, %r exists" % path
                return

        if not force and self.patch_egg(egg):
            return

        self.fetch(egg)


def main():
    from optparse import OptionParser
    import store.indexed as indexed
    from egg_meta import is_valid_eggname

    p = OptionParser(usage="usage: %prog [options] ROOT_URL [EGG ...]",
                     description="simple interface to fetch eggs")
    p.add_option("--auth",
                 action="store",
                 help="username:password")
    p.add_option("--dst",
                 action="store",
                 help="destination directory",
                 default=os.getcwd(),
                 metavar='PATH')
    p.add_option("--force",
                 action="store_true")
    p.add_option('-v', "--verbose", action="store_true")

    opts, args = p.parse_args()

    if len(args) < 1:
        p.error('at least one argument (the repo root URL) expected, try -h')

    repo_url = args[0]
    if repo_url.startswith(('http://', 'https://')):
        store = indexed.RemoteHTTPIndexedStore(repo_url)
    else:
        store = indexed.LocalIndexedStore(repo_url)

    store.connect(tuple(opts.auth.split(':', 1)) if opts.auth else None)

    f = FetchAPI(store, opts.dst)
    f.verbose = opts.verbose
    for fn in args[1:]:
        if not is_valid_eggname(fn):
            raise Exception('Error: invalid egg name: %r' % fn)
        f.fetch_egg(fn, opts.force)


if __name__ == '__main__':
    main()



import sys
import os
import shutil
import tempfile
from os.path import basename, isdir, isfile, islink, join


on_win = bool(sys.platform == 'win32')

if on_win:
    bin_dir_name = 'Scripts'
    rel_site_packages = r'Lib\site-packages'
else:
    bin_dir_name = 'bin'
    rel_site_packages = 'lib/python%i.%i/site-packages' % sys.version_info[:2]


def rm_empty_dir(path):
    """
    Remove the directory `path` if it is a directory and empty.
    If the directory does not exist or is not empty, do nothing.
    """
    try:
        os.rmdir(path)
    except OSError: # directory might not exist or not be empty
        pass


def rm_rf(path, verbose=False):
    if not on_win and islink(path):
        # Note that we have to check if the destination is a link because
        # exists('/path/to/dead-link') will return False, although
        # islink('/path/to/dead-link') is True.
        if verbose:
            print "Removing: %r (link)" % path
        os.unlink(path)

    elif isfile(path):
        if verbose:
            print "Removing: %r (file)" % path
        if on_win:
            try:
                os.unlink(path)
            except (WindowsError, IOError):
                os.rename(path, join(tempfile.mkdtemp(), basename(path)))
        else:
            os.unlink(path)

    elif isdir(path):
        if verbose:
            print "Removing: %r (directory)" % path
        if on_win:
            try:
                shutil.rmtree(path)
            except (WindowsError, IOError):
                os.rename(path, join(tempfile.mkdtemp(), basename(path)))
        else:
            shutil.rmtree(path)


def get_executable(prefix):
    if on_win:
        path = join(prefix, 'python.exe')
        if isfile(path):
            return path
    else:
        path = join(prefix, bin_dir_name, 'python')
        if isfile(path):
            from subprocess import Popen, PIPE
            cmd = [path, '-c', 'import sys;print sys.executable']
            p = Popen(cmd, stdout=PIPE)
            return p.communicate()[0].strip()
    return sys.executable


def human_bytes(n):
    """
    Return the number of bytes n in more human readable form.
    """
    if n < 1024:
        return '%i B' % n
    k = (n - 1) / 1024 + 1
    if k < 1024:
        return '%i KB' % k
    return '%.2f MB' % (float(n) / (2**20))



# unused
import os
import json
from os.path import isfile, join

from base import AbstractStore


class LocalStore(AbstractStore):

    def __init__(self, location):
        self.root = location
        self._index_path = join(self.root, 'index.json')

    def connect(self, auth=None):
        pass

    def get(self, key):
        return self.get_data(key), self.get_metadata(key)

    def get_data(self, key):
        try:
            return open(self.path(key), 'rb')
        except IOError as e:
            raise KeyError(str(e))

    def get_metadata(self, key):
        self._read_index()
        return self._index[key]

    def set(self, key, value):
        self.set_data(key, value[0])
        self.set_metadata(key, value[1])

    def set_data(self, key, value, buffer_size=1048576):
        with open(self.path(key), 'wb') as fo:
            while True:
                chunk = value.read(buffer_size)
                if not chunk:
                    break
                fo.write(chunk)

    def set_metadata(self, key, value):
        self._read_index()
        self._index[key] = value
        self._write_index()

    def delete(self, key):
        os.unlink(self.path(key))
        self._read_index()
        del self._index[key]
        self._write_index()

    def _read_index(self):
        if isfile(self._index_path):
            self._index = json.load(open(self._index_path))
        else:
            self._index = {}

    def _write_index(self):
        with open(self._index_path, 'w') as f:
            json.dump(self._index, f, indent=2, sort_keys=True)

    def exists(self, key):
        self._read_index()
        return key in self._index

    def query(self, **kwargs):
        res = {}
        for key in self.query_keys(**kwargs):
            res[key] = self._index[key]
        return res

    def query_keys(self, **kwargs):
        self._read_index()
        for key, info in self._index.iteritems():
            if all(info.get(k) == v for k, v in kwargs.iteritems()):
                yield key

    def path(self, key):
        return join(self.root, key)
